@@ -1,10 +1,13 @@
-from datetime import datetime
+from datetime import datetime, timedelta
 from abc import abstractmethod
+from typing import Optional
 
 import warnings
 import numpy as np
 import pandas as pd
 from tqdm import tqdm, trange
+
+from mikeio.spatial.geometry import GeometryUndefined
 from .dataset import Dataset
 from .base import TimeSeries
 
@@ -12,9 +15,35 @@ from .dfsutil import _valid_item_numbers, _valid_timesteps, _get_item_info
 from .eum import ItemInfo, TimeStepUnit, EUMType, EUMUnit
 from .custom_exceptions import DataDimensionMismatch, ItemNumbersError
 from mikecore.eum import eumQuantity
-from mikecore.DfsFile import DfsSimpleType, TimeAxisType
+from mikecore.DfsFile import DfsSimpleType, TimeAxisType, DfsFile
 from mikecore.DfsFactory import DfsFactory
 
+def _write_dfs_data(*, dfs: DfsFile, ds: Dataset, n_spatial_dims: int) -> None:
+
+    deletevalue = dfs.FileInfo.DeleteValueFloat  # ds.deletevalue
+    has_no_time = "time" not in ds.dims
+    if ds.is_equidistant:
+        t_rel = np.zeros(ds.n_timesteps)
+    else:
+        t_rel = (ds.time - ds.time[0]).total_seconds()
+    
+    for i in range(ds.n_timesteps):
+        for item in range(ds.n_items):
+
+            if has_no_time:
+                d = ds[item].values
+            else:
+                d = ds[item].values[i]
+            d = d.copy()  # to avoid modifying the input
+            d[np.isnan(d)] = deletevalue
+
+            d = d.reshape(ds.shape[-n_spatial_dims:])  # spatial axes
+            darray = d.flatten()
+
+            dfs.WriteItemTimeStepNext(t_rel[i], darray.astype(np.float32))
+
+    dfs.Close()
+
 
 class _Dfs123(TimeSeries):
 
@@ -33,8 +62,18 @@ class _Dfs123(TimeSeries):
         self._override_coordinates = False
         self._timeseries_unit = TimeStepUnit.SECOND
         self._dt = None
+        self.geometry = GeometryUndefined()
+        self._dfs = None
+        self._source = None
 
-    def read(self, items=None, time_steps=None):
+    def read(
+        self,
+        *,
+        items=None,
+        time=None,
+        keepdims=False,
+        dtype=np.float32,
+    ) -> Dataset:
         """
         Read data from a dfs file
 
@@ -42,20 +81,24 @@ class _Dfs123(TimeSeries):
         ---------
         items: list[int] or list[str], optional
             Read only selected items, by number (0-based), or by name
-        time_steps: str, int or list[int], optional
-            Read only selected time_steps
+        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
+            Read only selected time steps, by default None (=all)
+        keepdims: bool, optional
+            When reading a single time step only, should the time-dimension be kept
+            in the returned Dataset? by default: False
 
         Returns
         -------
         Dataset
         """
+
         self._open()
 
         item_numbers = _valid_item_numbers(self._dfs.ItemInfo, items)
         n_items = len(item_numbers)
 
-        time_steps = _valid_timesteps(self._dfs.FileInfo, time_steps)
-        nt = len(time_steps)
+        single_time_selected, time_steps = _valid_timesteps(self._dfs.FileInfo, time)
+        nt = len(time_steps) if not single_time_selected else 1
 
         if self._ndim == 1:
             shape = (nt, self._nx)
@@ -64,7 +107,10 @@ class _Dfs123(TimeSeries):
         else:
             shape = (nt, self._nz, self._ny, self._nx)
 
-        data_list = [np.ndarray(shape=shape) for item in range(n_items)]
+        if single_time_selected and not keepdims:
+            shape = shape[1:]
+
+        data_list = [np.ndarray(shape=shape, dtype=dtype) for item in range(n_items)]
 
         t_seconds = np.zeros(len(time_steps))
 
@@ -80,9 +126,11 @@ class _Dfs123(TimeSeries):
 
                 if self._ndim == 2:
                     d = d.reshape(self._ny, self._nx)
-                    d = np.flipud(d)
 
-                data_list[item][i] = d
+                if single_time_selected:
+                    data_list[item] = d
+                else:
+                    data_list[item][i] = d
 
             t_seconds[i] = itemdata.Time
 
@@ -91,7 +139,7 @@ class _Dfs123(TimeSeries):
         items = _get_item_info(self._dfs.ItemInfo, item_numbers)
 
         self._dfs.Close()
-        return Dataset(data_list, time, items)
+        return Dataset(data_list, time, items, geometry=self.geometry, validate=False)
 
     def _read_header(self):
         dfs = self._dfs
@@ -111,7 +159,8 @@ class _Dfs123(TimeSeries):
             )  # TODO handle other timeunits
             # TODO to get the EndTime
         self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps
-        self._projstr = dfs.FileInfo.Projection.WKTString
+        projstr = dfs.FileInfo.Projection.WKTString
+        self._projstr = "NON-UTM" if not projstr else projstr
         self._longitude = dfs.FileInfo.Projection.Longitude
         self._latitude = dfs.FileInfo.Projection.Latitude
         self._orientation = dfs.FileInfo.Projection.Orientation
@@ -121,47 +170,52 @@ class _Dfs123(TimeSeries):
 
     def _write(
         self,
+        *,
         filename,
         data,
-        start_time,
         dt,
-        datetimes,
-        items,
-        coordinate,
+        coordinate=None,
         title,
         keep_open=False,
     ):
 
+        neq_datetimes = None
         if isinstance(data, Dataset) and not data.is_equidistant:
-            datetimes = data.time
+            neq_datetimes = data.time
 
         self._write_handle_common_arguments(
-            title, data, items, coordinate, start_time, dt
+            title=title, data=data, coordinate=coordinate, dt=dt
         )
 
         shape = np.shape(data[0])
+        t_offset = 0 if len(shape) == self._ndim else 1
         if self._ndim == 1:
-            self._nx = shape[1]
+            self._nx = shape[t_offset + 0]
         elif self._ndim == 2:
-            self._ny = shape[1]
-            self._nx = shape[2]
+            self._ny = shape[t_offset + 0]
+            self._nx = shape[t_offset + 1]
+        elif self._ndim == 3:
+            self._nz = shape[t_offset + 0]
+            self._ny = shape[t_offset + 1]
+            self._nx = shape[t_offset + 2]
 
         self._factory = DfsFactory()
         self._set_spatial_axis()
 
         if self._ndim == 1:
-            if not all(np.shape(d)[1] == self._nx for d in data):
+            if not all(np.shape(d)[t_offset + 0] == self._nx for d in self._data):
                 raise DataDimensionMismatch()
 
         if self._ndim == 2:
-            if not all(np.shape(d)[1] == self._ny for d in data):
+            if not all(np.shape(d)[t_offset + 0] == self._ny for d in self._data):
                 raise DataDimensionMismatch()
 
-            if not all(np.shape(d)[2] == self._nx for d in data):
+            if not all(np.shape(d)[t_offset + 1] == self._nx for d in self._data):
                 raise DataDimensionMismatch()
-        if datetimes is not None:
+
+        if neq_datetimes is not None:
             self._is_equidistant = False
-            start_time = datetimes[0]
+            start_time = neq_datetimes[0]
             self._start_time = start_time
 
         dfs = self._setup_header(filename)
@@ -172,36 +226,28 @@ class _Dfs123(TimeSeries):
         for i in trange(self._n_timesteps, disable=not self.show_progress):
             for item in range(self._n_items):
 
-                d = self._data[item][i]
+                d = self._data[item][i] if t_offset == 1 else self._data[item]
                 d = d.copy()  # to avoid modifying the input
                 d[np.isnan(d)] = deletevalue
 
-                if self._ndim == 1:
-                    darray = d
-
-                if self._ndim == 2:
-                    d = d.reshape(self.shape[1:])
-                    d = np.flipud(d)
-                    darray = d.reshape(d.size, 1)[:, 0]
-
                 if self._is_equidistant:
-                    dfs.WriteItemTimeStepNext(0, darray.astype(np.float32))
+                    dfs.WriteItemTimeStepNext(0, d.astype(np.float32))
                 else:
-                    t = datetimes[i]
+                    t = neq_datetimes[i]
                     relt = (t - self._start_time).total_seconds()
-                    dfs.WriteItemTimeStepNext(relt, darray.astype(np.float32))
+                    dfs.WriteItemTimeStepNext(relt, d.astype(np.float32))
 
         if not keep_open:
             dfs.Close()
         else:
             return self
 
-    def append(self, data):
+    def append(self, data: Dataset) -> None:
         """Append to a dfs file opened with `write(...,keep_open=True)`
 
         Parameters
         -----------
-        data: list[np.array]
+        data: Dataset
         """
 
         deletevalue = self._dfs.FileInfo.DeleteValueFloat  # -1.0000000031710769e-30
@@ -209,7 +255,7 @@ class _Dfs123(TimeSeries):
         for i in trange(self._n_timesteps, disable=not self.show_progress):
             for item in range(self._n_items):
 
-                d = data[item][i]
+                d = data[item].to_numpy()[i]
                 d = d.copy()  # to avoid modifying the input
                 d[np.isnan(d)] = deletevalue
 
@@ -218,7 +264,6 @@ class _Dfs123(TimeSeries):
 
                 if self._ndim == 2:
                     d = d.reshape(self.shape[1:])
-                    d = np.flipud(d)
                     darray = d.reshape(d.size, 1)[:, 0]
 
                 if self._is_equidistant:
@@ -238,9 +283,7 @@ class _Dfs123(TimeSeries):
         "Finalize write for a dfs file opened with `write(...,keep_open=True)`"
         self._dfs.Close()
 
-    def _write_handle_common_arguments(
-        self, title, data, items, coordinate, start_time, dt
-    ):
+    def _write_handle_common_arguments(self, *, title, data, coordinate, dt):
 
         if title is None:
             self._title = ""
@@ -256,6 +299,13 @@ class _Dfs123(TimeSeries):
                     self._latitude,
                     self._orientation,
                 ]
+            elif isinstance(data, Dataset) and (data.geometry is not None):
+                self._coordinate = [
+                    data.geometry.projection_string,
+                    data.geometry.origin[0],
+                    data.geometry.origin[1],
+                    data.geometry.orientation,
+                ]
             else:
                 warnings.warn("No coordinate system provided")
                 self._coordinate = ["LONG/LAT", 0, 0, 0]
@@ -266,22 +316,12 @@ class _Dfs123(TimeSeries):
         if isinstance(data, Dataset):
             self._items = data.items
             self._start_time = data.time[0]
+            self._n_timesteps = len(data.time)
             if dt is None and len(data.time) > 1:
                 self._dt = (data.time[1] - data.time[0]).total_seconds()
-            self._data = data.data
+            self._data = data.to_numpy()
         else:
-            self._data = data
-
-        if start_time is None:
-            if self._start_time is None:
-                self._start_time = datetime.now()
-                warnings.warn(
-                    f"No start time supplied. Using current time: {self._start_time} as start time."
-                )
-            else:
-                self._start_time = self._start_time
-        else:
-            self._start_time = start_time
+            raise TypeError("data must be supplied in the form of a mikeio.Dataset")
 
         if dt:
             self._dt = dt
@@ -291,9 +331,6 @@ class _Dfs123(TimeSeries):
             if self._n_timesteps > 1:
                 warnings.warn("No timestep supplied. Using 1s.")
 
-        if items:
-            self._items = items
-
         if self._items is None:
             self._items = [ItemInfo(f"Item {i+1}") for i in range(self._n_items)]
 
@@ -428,19 +465,40 @@ class _Dfs123(TimeSeries):
     def end_time(self):
         """File end time"""
         if self._end_time is None:
-            self._end_time = self.read([0]).time[-1].to_pydatetime()
+            self._end_time = self.read(items=[0]).time[-1].to_pydatetime()
 
         return self._end_time
 
     @property
-    def n_timesteps(self):
+    def n_timesteps(self) -> int:
         """Number of time steps"""
         return self._n_timesteps
 
     @property
-    def timestep(self):
+    def timestep(self) -> Optional[float]:
         """Time step size in seconds"""
-        return self._timestep_in_seconds
+        if self._timeaxistype in {
+            TimeAxisType.CalendarEquidistant,
+            TimeAxisType.TimeEquidistant,
+        }:
+            return self._dfs.FileInfo.TimeAxis.TimeStepInSeconds()
+
+    @property
+    def time(self) -> Optional[pd.DatetimeIndex]:
+        """File all datetimes"""
+        if self._timeaxistype in {
+            TimeAxisType.CalendarEquidistant,
+            TimeAxisType.TimeEquidistant,
+        }:
+            return pd.to_datetime(
+                [
+                    self.start_time + timedelta(seconds=i * self.timestep)
+                    for i in range(self.n_timesteps)
+                ]
+            )
+
+        else:
+            return None
 
     @property
     def projection_string(self):
@@ -456,10 +514,15 @@ class _Dfs123(TimeSeries):
         """Origin latitude"""
         return self._latitude
 
+    @property
+    def origin(self):
+        """Origin (in own CRS)"""
+        return self.geometry.origin
+
     @property
     def orientation(self):
-        """North to Y orientation"""
-        return self._orientation
+        """Orientation (in own CRS)"""
+        return self.geometry.orientation
 
     @property
     @abstractmethod
