@@ -16,12 +16,78 @@ from mikecore.DfsFile import (
     TimeAxisType,
 )
 
-from .dfsutil import _valid_item_numbers, _get_item_info
+from . import __dfs_version__
+from .dfsutil import _valid_item_numbers, _get_item_info, _valid_timesteps
 from .dataset import Dataset
 from .eum import TimeStepUnit, EUMType, EUMUnit, ItemInfo
 from .base import TimeSeries
 
 
+def _write_dfs0(filename, dataset: Dataset, title="", dtype=DfsSimpleType.Float):
+    filename = str(filename)
+
+    factory = DfsFactory()
+    builder = DfsBuilder.Create(title, "mikeio", __dfs_version__)
+    builder.SetDataType(1)
+    builder.SetGeographicalProjection(factory.CreateProjectionUndefined())
+
+    system_start_time = dataset.time[0]
+
+    if dataset.is_equidistant:
+        if len(dataset.time) == 1:
+            dt = 1.0  # TODO
+        else:
+            dt = (dataset.time[1] - dataset.time[0]).total_seconds()
+
+        temporal_axis = factory.CreateTemporalEqCalendarAxis(
+            TimeStepUnit.SECOND, system_start_time, 0, dt
+        )
+    else:
+        temporal_axis = factory.CreateTemporalNonEqCalendarAxis(
+            TimeStepUnit.SECOND, system_start_time
+        )
+
+    builder.SetTemporalAxis(temporal_axis)
+    builder.SetItemStatisticsType(StatType.RegularStat)
+
+    dfs_dtype = Dfs0._to_dfs_datatype(dtype)
+
+    for da in dataset:
+        newitem = builder.CreateDynamicItemBuilder()
+        quantity = eumQuantity.Create(da.type, da.unit)
+        newitem.Set(da.name, quantity, dfs_dtype)
+
+        # TODO set default on DataArray
+        if da.item.data_value_type is not None:
+            newitem.SetValueType(da.item.data_value_type)
+        else:
+            newitem.SetValueType(DataValueType.Instantaneous)
+
+        newitem.SetAxis(factory.CreateAxisEqD0())
+        builder.AddDynamicItem(newitem.GetDynamicItemInfo())
+
+    builder.CreateFile(filename)
+
+    dfs = builder.GetFile()
+
+    delete_value = dfs.FileInfo.DeleteValueFloat
+
+    t_seconds = (dataset.time - dataset.time[0]).total_seconds().values
+
+    data = np.array(dataset.to_numpy(), order="F").astype(np.float64).T
+    data[np.isnan(data)] = delete_value
+
+    if data.ndim == 2:
+        data_to_write = np.concatenate([t_seconds.reshape(-1, 1), data], axis=1)
+    else:
+        data_to_write = np.concatenate(
+            [np.atleast_2d(t_seconds), np.atleast_2d(data)], axis=1
+        )
+    rc = dfs.WriteDfs0DataDouble(data_to_write)
+
+    dfs.Close()
+
+
 class Dfs0(TimeSeries):
     def __init__(self, filename=None):
         """Create a Dfs0 object for reading, writing
@@ -42,7 +108,6 @@ class Dfs0(TimeSeries):
         self._title = None
         self._items = None
         self._n_timesteps = None
-        self._dtype = None
 
         self._filename = str(filename)
 
@@ -53,15 +118,15 @@ class Dfs0(TimeSeries):
         out = ["<mikeio.Dfs0>"]
 
         if os.path.isfile(self._filename):
-            out.append(f"Timeaxis: {str(self._timeaxistype)}")
+            out.append(f"timeaxis: {repr(self._timeaxistype)}")
 
         if self._n_items is not None:
             if self._n_items < 10:
-                out.append("Items:")
+                out.append("items:")
                 for i, item in enumerate(self.items):
                     out.append(f"  {i}:  {item}")
             else:
-                out.append(f"Number of items: {self._n_items}")
+                out.append(f"number of items: {self._n_items}")
 
         return str.join("\n", out)
 
@@ -87,9 +152,12 @@ class Dfs0(TimeSeries):
         else:  # relative time axis
             self._start_time = datetime(1970, 1, 1)
 
+        # time
+        self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps
+
         dfs.Close()
 
-    def read(self, items=None, time_steps=None):
+    def read(self, items=None, time=None, keepdims=False) -> Dataset:
         """
         Read data from a dfs0 file.
 
@@ -97,8 +165,8 @@ class Dfs0(TimeSeries):
         ----------
         items: list[int] or list[str], optional
             Read only selected items, by number (0-based), or by name
-        time_steps: str, int or list[int], optional
-            Read only selected time_steps
+        time: int, str, datetime, pd.TimeStamp, sequence, slice or pd.DatetimeIndex, optional
+            Read only selected time steps, by default None (=all)
 
         Returns
         -------
@@ -109,27 +177,32 @@ class Dfs0(TimeSeries):
         if not os.path.exists(self._filename):
             raise FileNotFoundError(f"File {self._filename} not found.")
 
-        dfs = DfsFileFactory.DfsGenericOpen(self._filename)
-        self._source = dfs
-        self._dfs = dfs
+        # read data from file
+        fdata, ftime, fitems = self.__read(self._filename)
+        self._source = self._dfs
+        dfs = self._dfs
 
+        # select items
         self._n_items = len(dfs.ItemInfo)
         item_numbers = _valid_item_numbers(dfs.ItemInfo, items)
+        if items is not None:
+            fdata = [fdata[it] for it in item_numbers]
+            fitems = [fitems[it] for it in item_numbers]
+        ds = Dataset(fdata, ftime, fitems, validate=False)
 
+        # select time steps
         self._n_timesteps = dfs.FileInfo.TimeAxis.NumberOfTimeSteps
-
         if self._timeaxistype == TimeAxisType.CalendarNonEquidistant and isinstance(
-            time_steps, str
+            time, str
         ):
-            sel_time_step_str = time_steps
-            time_steps = range(self._n_timesteps)
+            sel_time_step_str = time
+            time_steps = None
         else:
             sel_time_step_str = None
+            time_steps = None
+            if time is not None:
+                _, time_steps = _valid_timesteps(dfs.FileInfo, time)
 
-        dfs.Close()
-
-        ds = self.__read(self._filename)
-        ds = ds[item_numbers]
         if time_steps:
             ds = ds.isel(time_steps, axis=0)
 
@@ -179,7 +252,7 @@ class Dfs0(TimeSeries):
             for item in self._dfs.ItemInfo
         ]
 
-        return Dataset(data, time, items)
+        return data, time, items
 
     @staticmethod
     def _to_dfs_datatype(dtype):
@@ -196,7 +269,7 @@ class Dfs0(TimeSeries):
 
     def _setup_header(self):
         factory = DfsFactory()
-        builder = DfsBuilder.Create(self._title, "DFS", 0)
+        builder = DfsBuilder.Create(self._title, "mikeio", __dfs_version__)
         builder.SetDataType(1)
         builder.SetGeographicalProjection(factory.CreateProjectionUndefined())
 
@@ -246,7 +319,6 @@ class Dfs0(TimeSeries):
         filename,
         data,
         start_time=None,
-        timeseries_unit=TimeStepUnit.SECOND,
         dt=None,
         datetimes=None,
         items=None,
@@ -288,7 +360,7 @@ class Dfs0(TimeSeries):
                 self._dt = (data.time[1] - data.time[0]).total_seconds()
             else:
                 datetimes = data.time
-            data = data.data
+            data = data.to_numpy()
         elif datetimes is not None:
             datetimes = pd.DatetimeIndex(datetimes, freq="infer")
 
@@ -344,7 +416,7 @@ class Dfs0(TimeSeries):
 
         dfs.Close()
 
-    def to_dataframe(self, unit_in_name=False, round_time="ms"):
+    def to_dataframe(self, unit_in_name=False, round_time="ms") -> pd.DataFrame:
         """
         Read data from the dfs0 file and return a Pandas DataFrame.
 
@@ -358,8 +430,18 @@ class Dfs0(TimeSeries):
         -------
         pd.DataFrame
         """
-        ds = self.read()
-        df = ds.to_dataframe(unit_in_name=unit_in_name, round_time=round_time)
+        data, time, items = self.__read(self._filename)
+        if unit_in_name:
+            cols = [f"{item.name} ({item.unit.name})" for item in items]
+        else:
+            cols = [f"{item.name}" for item in items]
+        df = pd.DataFrame(np.atleast_2d(data).T, index=time, columns=cols)
+
+        if round_time:
+            rounded_idx = pd.DatetimeIndex(time).round(round_time)
+            df.index = pd.DatetimeIndex(rounded_idx, freq="infer")
+        else:
+            df.index = pd.DatetimeIndex(time, freq="infer")
 
         return df
 
@@ -423,6 +505,35 @@ class Dfs0(TimeSeries):
         """Number of time steps"""
         return self._n_timesteps
 
+    @property
+    def timestep(self):
+        """Time step size in seconds"""
+        if self._timeaxistype == TimeAxisType.CalendarEquidistant:
+            return self._source.FileInfo.TimeAxis.TimeStep
+
+    @property
+    def time(self):
+        """File all datetimes"""
+        if self._timeaxistype == TimeAxisType.CalendarEquidistant:
+            return pd.to_datetime(
+                [
+                    self.start_time + timedelta(seconds=i * self.timestep)
+                    for i in range(self.n_timesteps)
+                ]
+            )
+
+        elif self._timeaxistype == TimeAxisType.CalendarNonEquidistant:
+            dfs = DfsFileFactory.DfsGenericOpen(self._filename)
+            t_seconds = np.zeros(self.n_timesteps)
+            for it in range(self.n_timesteps):
+                itemdata = dfs.ReadItemTimeStep(1, int(it))
+                t_seconds[it] = itemdata.Time
+
+            return pd.to_datetime(t_seconds, unit="s", origin=self.start_time)
+
+        else:
+            return None
+
 
 def series_to_dfs0(
     self,
@@ -514,27 +625,3 @@ def dataframe_to_dfs0(
 pd.DataFrame.to_dfs0 = dataframe_to_dfs0
 
 pd.Series.to_dfs0 = series_to_dfs0
-
-
-def dataset_to_dfs0(self, filename):
-    """Write Dataset to a Dfs0 file
-
-    Parameters
-    ----------
-    filename: str
-        full path and file name to the dfs0 file.
-    """
-    self = self.squeeze()
-
-    if len(self.data[0].shape) != 1:
-        raise ValueError(
-            """Only data with a single dimension can be converted to a dfs0.
-                 Hint: use `isel` to create a subset."""
-        )
-
-    dfs0 = Dfs0()
-
-    dfs0.write(filename, self)
-
-
-Dataset.to_dfs0 = dataset_to_dfs0
