@@ -1,11 +1,17 @@
 import os
 import warnings
+import numpy as np
+import pandas as pd
+from datetime import timedelta
 
 from mikecore.eum import eumUnit
 from mikecore.DfsFileFactory import DfsFileFactory
 from mikecore.DfsBuilder import DfsBuilder
 
+from . import __dfs_version__
 from .dfs import _Dfs123
+from .spatial.grid_geometry import Grid1D
+from mikecore.DfsFile import TimeAxisType
 
 
 class Dfs1(_Dfs123):
@@ -20,6 +26,15 @@ class Dfs1(_Dfs123):
 
         if filename:
             self._read_dfs1_header()
+            origin = self._longitude, self._latitude
+            self.geometry = Grid1D(
+                x0=self._x0,
+                dx=self._dx,
+                nx=self._nx,
+                projection=self._projstr,
+                origin=origin,
+                orientation=self._orientation,
+            )
 
     def __repr__(self):
         out = ["<mikeio.Dfs1>"]
@@ -29,17 +44,17 @@ class Dfs1(_Dfs123):
 
             if self._n_items is not None:
                 if self._n_items < 10:
-                    out.append("Items:")
+                    out.append("items:")
                     for i, item in enumerate(self.items):
                         out.append(f"  {i}:  {item}")
                 else:
-                    out.append(f"Number of items: {self._n_items}")
+                    out.append(f"number of items: {self._n_items}")
             if os.path.isfile(self._filename):
                 if self._n_timesteps == 1:
-                    out.append("Time: time-invariant file (1 step)")
+                    out.append("time: time-invariant file (1 step)")
                 else:
-                    out.append(f"Time: {self._n_timesteps} steps")
-                    out.append(f"Start time: {self._start_time}")
+                    out.append(f"time: {self._n_timesteps} steps")
+                    out.append(f"start time: {self._start_time}")
 
         return str.join("\n", out)
 
@@ -48,6 +63,7 @@ class Dfs1(_Dfs123):
             raise FileNotFoundError(self._filename)
 
         self._dfs = DfsFileFactory.Dfs1FileOpen(self._filename)
+        self._x0 = self._dfs.SpatialAxis.X0
         self._dx = self._dfs.SpatialAxis.Dx
         self._nx = self._dfs.SpatialAxis.XCount
 
@@ -60,13 +76,9 @@ class Dfs1(_Dfs123):
         self,
         filename,
         data,
-        start_time=None,
         dt=None,
-        datetimes=None,
-        items=None,
         dx=1,
         x0=0,
-        coordinate=None,
         title=None,
     ):
         """
@@ -76,19 +88,10 @@ class Dfs1(_Dfs123):
         ----------
         filename: str
             Location to write the dfs1 file
-        data: list[np.array]
+        data: Dataset
             list of matrices, one for each item. Matrix dimension: x, time
-        start_time: datetime, optional
-            start datetime
         dt: float
             The time step in seconds.
-        datetimes: List[datetime], optional, deprecated
-            The list of datetimes for the case of non-equidistant Timeaxis.
-        items: list[ItemInfo], optional
-            List of ItemInfo (e.g. Water Level).
-        coordinate:
-            list of [projection, origin_x, origin_y, orientation]
-            e.g. ['LONG/LAT', 12.4387, 55.2257, 327]
         x0:
             Lower right position
         dx:
@@ -100,33 +103,14 @@ class Dfs1(_Dfs123):
 
         self._x0 = x0
 
-        if start_time:
-            warnings.warn(
-                "setting start_time is deprecated, please supply data in the form of a Dataset",
-                FutureWarning,
-            )
-
-        if datetimes:
-            warnings.warn(
-                "setting datetimes is deprecated, please supply data in the form of a Dataset",
-                FutureWarning,
-            )
-
-        if items:
-            warnings.warn(
-                "setting items is deprecated, please supply data in the form of a Dataset",
-                FutureWarning,
-            )
-
         if isinstance(data, list):
-            warnings.warn(
-                "supplying data as a list of numpy arrays is deprecated, please supply data in the form of a Dataset",
-                FutureWarning,
+            raise TypeError(
+                "supplying data as a list of numpy arrays is deprecated, please supply data in the form of a Dataset"
             )
 
-        self._builder = DfsBuilder.Create(title, "mikeio", 0)
+        self._builder = DfsBuilder.Create(title, "mikeio", __dfs_version__)
         self._dx = dx
-        self._write(filename, data, start_time, dt, datetimes, items, coordinate, title)
+        self._write(filename=filename, data=data, dt=dt, title=title)
 
     def _set_spatial_axis(self):
         self._builder.SetSpatialAxis(
@@ -135,7 +119,17 @@ class Dfs1(_Dfs123):
             )
         )
 
+    @property
+    def x0(self):
+        """Start point of x values (often 0)"""
+        return self._x0
+
     @property
     def dx(self):
         """Step size in x direction"""
         return self._dx
+
+    @property
+    def nx(self):
+        """Number of node values"""
+        return self._nx
