@@ -12,13 +12,15 @@ Examples
 degree Kelvin
 
 """
-from typing import List
+from typing import List, Sequence
 from mikecore.DfsFile import DataValueType
 from mikecore.eum import eumWrapper
 from enum import IntEnum
 
 from mikeio.helpers import to_datatype
 
+import pandas as pd
+
 
 def type_list(search=None):
     """Get a dictionary of the EUM items
@@ -703,6 +705,9 @@ class EUMType(IntEnum):
     NearBedLoad_per_Length = 110307
     Substance_per_UnitArea = 110308
     AccNearBedLoad_per_Length = 110309
+    ThermalConductivity = 110310
+    DirectionalVariance = 110311
+    SpecificDissipationRate = 110312
 
     def __init__(self, code):
         self.code = code
@@ -1160,6 +1165,7 @@ class EUMUnit(IntEnum):
     pound_per_Feet_per_Second = 2312
     degree50 = 2402
     degree_pow_2 = 2403
+    radian_pow_2 = 2404
     degree_per_meter = 2500
     radian_per_meter = 2501
     degree_per_sec = 2510
@@ -1331,6 +1337,8 @@ class EUMUnit(IntEnum):
     MW_per_meter = 99202
     GW_per_meter = 99203
     kW_per_feet = 99204
+    watt_per_meter_per_degree_celsius = 99220
+    watt_per_feet_per_degree_fahrenheit = 99221
     siemens = 99250
     millisiemens = 99251
     microsiemens = 99252
@@ -1448,3 +1456,12 @@ class ItemInfo:
             return f"{self.name} <{self.type.display_name}> ({self.unit.display_name})"
         else:
             return f"{self.name} <{self.type.display_name}> ({self.unit.display_name}) - {self.data_value_type}"
+
+class ItemInfoList(list):
+
+    def __init__(self, items: Sequence[ItemInfo]):
+        super().__init__(items)
+
+    def to_dataframe(self):
+        data = [{"name": item.name, "type": item.type.name, "unit": item.unit.name} for item in self]
+        return pd.DataFrame(data)
