@@ -14,9 +14,11 @@ from mikecore.DfsFileFactory import DfsFileFactory
 from mikecore.DfsBuilder import DfsBuilder
 from mikecore.DfsFile import DfsDynamicItemInfo, DfsFile
 from mikecore.eum import eumQuantity
+from . import __dfs_version__
 from .dfsutil import _valid_item_numbers, _get_item_info
 from .eum import EUMType, ItemInfo
 
+
 show_progress = False
 
 
@@ -57,7 +59,9 @@ class _ChunkInfo:
         return _ChunkInfo(n_data, n_chunks)
 
 
-def _clone(infilename: str, outfilename: str, start_time=None, items=None) -> DfsFile:
+def _clone(
+    infilename: str, outfilename: str, start_time=None, timestep=None, items=None
+) -> DfsFile:
     """Clone a dfs file
 
     Parameters
@@ -67,7 +71,9 @@ def _clone(infilename: str, outfilename: str, start_time=None, items=None) -> Df
     outfilename : str
         output filename
     start_time : datetime, optional
-        new start time for the new file, default
+        new start time for the new file, by default None
+    timestep : float, optional
+        new timestep (in seconds) for the new file, by default None
     items : list(int,str,eum.ItemInfo), optional
         list of items for new file, either as a list of
         ItemInfo or a list of str/int referring to original file,
@@ -81,9 +87,7 @@ def _clone(infilename: str, outfilename: str, start_time=None, items=None) -> Df
     source = DfsFileFactory.DfsGenericOpen(infilename)
     fi = source.FileInfo
 
-    builder = DfsBuilder.Create(
-        fi.FileTitle, fi.ApplicationTitle, fi.ApplicationVersion
-    )
+    builder = DfsBuilder.Create(fi.FileTitle, "mikeio", __dfs_version__)
 
     # Set up the header
     builder.SetDataType(fi.DataType)
@@ -93,6 +97,8 @@ def _clone(infilename: str, outfilename: str, start_time=None, items=None) -> Df
     time_axis = fi.TimeAxis
     if start_time is not None:
         time_axis.StartDateTime = start_time
+    if timestep is not None:
+        time_axis.TimeStep = timestep
     builder.SetTemporalAxis(time_axis)
 
     builder.SetItemStatisticsType(fi.StatsType)
@@ -289,16 +295,17 @@ def diff(infilename_a: str, infilename_b: str, outfilename: str) -> None:
 
             outdata = d_a - d_b
 
-            darray = outdata.astype(np.float32)
+            d = outdata.astype(np.float32)
+            d[np.isnan(d)] = deletevalue
 
-            dfs_o.WriteItemTimeStep(item + 1, timestep, time, darray)
+            dfs_o.WriteItemTimeStep(item + 1, timestep, time, d)
 
     dfs_i_a.Close()
     dfs_i_b.Close()
     dfs_o.Close()
 
 
-def concat(infilenames: List[str], outfilename: str) -> None:
+def concat(infilenames: List[str], outfilename: str, keep="last") -> None:
     """Concatenates files along the time axis
 
     If files are overlapping, the last one will be used.
@@ -309,6 +316,9 @@ def concat(infilenames: List[str], outfilename: str) -> None:
         filenames to concatenate
     outfilename: str
         filename of output
+    keep: str
+        either 'first' (keep older), 'last' (keep newer)
+        or 'average' can be selected. By default 'last'
 
     Notes
     ------
@@ -316,9 +326,9 @@ def concat(infilenames: List[str], outfilename: str) -> None:
     The list of input files have to be sorted, i.e. in chronological order
     """
 
-    dfs_i_a = DfsFileFactory.DfsGenericOpen(infilenames[0])
+    dfs_i_a = DfsFileFactory.DfsGenericOpen(str(infilenames[0]))
 
-    dfs_o = _clone(infilenames[0], outfilename)
+    dfs_o = _clone(str(infilenames[0]), str(outfilename))
 
     n_items = len(dfs_i_a.ItemInfo)
     dfs_i_a.Close()
@@ -327,7 +337,7 @@ def concat(infilenames: List[str], outfilename: str) -> None:
 
     for i, infilename in enumerate(tqdm(infilenames, disable=not show_progress)):
 
-        dfs_i = DfsFileFactory.DfsGenericOpen(infilename)
+        dfs_i = DfsFileFactory.DfsGenericOpen(str(infilename))
         t_axis = dfs_i.FileInfo.TimeAxis
         n_time_steps = t_axis.NumberOfTimeSteps
         dt = t_axis.TimeStep
@@ -340,35 +350,82 @@ def concat(infilenames: List[str], outfilename: str) -> None:
 
         current_time = start_time
 
-        if i < (len(infilenames) - 1):
-            dfs_n = DfsFileFactory.DfsGenericOpen(infilenames[i + 1])
-            nf = dfs_n.FileInfo.TimeAxis.StartDateTime
-            next_start_time = datetime(
-                nf.year, nf.month, nf.day, nf.hour, nf.minute, nf.second
-            )
-            dfs_n.Close()
+        if keep == "last":
 
-        for timestep in range(n_time_steps):
-
-            current_time = start_time + timedelta(seconds=timestep * dt)
             if i < (len(infilenames) - 1):
-                if current_time >= next_start_time:
-                    break
+                dfs_n = DfsFileFactory.DfsGenericOpen(str(infilenames[i + 1]))
+                nf = dfs_n.FileInfo.TimeAxis.StartDateTime
+                next_start_time = datetime(
+                    nf.year, nf.month, nf.day, nf.hour, nf.minute, nf.second
+                )
+                dfs_n.Close()
 
-            for item in range(n_items):
+            for timestep in range(n_time_steps):
 
-                itemdata = dfs_i.ReadItemTimeStep(item + 1, timestep)
-                d = itemdata.Data
+                current_time = start_time + timedelta(seconds=timestep * dt)
+                if i < (len(infilenames) - 1):
+                    if current_time >= next_start_time:
+                        break
 
-                darray = d.astype(np.float32)
+                for item in range(n_items):
 
-                dfs_o.WriteItemTimeStepNext(0, darray)
-        dfs_i.Close()
+                    itemdata = dfs_i.ReadItemTimeStep(item + 1, timestep)
+                    d = itemdata.Data
+
+                    darray = d.astype(np.float32)
+
+                    dfs_o.WriteItemTimeStepNext(0, darray)
+            dfs_i.Close()
+
+        if keep == "first":
+
+            if (
+                i == 0
+            ):  # all timesteps in first file are kept (is there a more efficient way to do this without the loop?)
+                for timestep in range(n_time_steps):
+                    current_time = start_time + timedelta(seconds=timestep * dt)
+
+                    for item in range(n_items):
+
+                        itemdata = dfs_i.ReadItemTimeStep(item + 1, timestep)
+                        d = itemdata.Data
+
+                        darray = d.astype(np.float32)
+
+                        dfs_o.WriteItemTimeStepNext(0, darray)
+                end_time = start_time + timedelta(
+                    seconds=timestep * dt
+                )  # reuse last timestep since there is no EndDateTime attribute in t_axis.
+                dfs_i.Close()
+
+            else:
+                # determine overlap in number of timesteps
+                start_timestep = (
+                    int((end_time - start_time) / timedelta(seconds=dt)) + 1
+                )
+                # loop only through those timesteps that are not in previous file
+                for timestep in range(start_timestep, n_time_steps):
+                    current_time = start_time + timedelta(seconds=timestep * dt)
+
+                    for item in range(n_items):
+
+                        itemdata = dfs_i.ReadItemTimeStep(item + 1, timestep)
+                        d = itemdata.Data
+
+                        darray = d.astype(np.float32)
+
+                        dfs_o.WriteItemTimeStepNext(0, darray)
+                end_time = start_time + timedelta(
+                    seconds=timestep * dt
+                )  # get end time from current file
+                dfs_i.Close()
 
     dfs_o.Close()
 
 
-def extract(infilename: str, outfilename: str, start=0, end=-1, items=None) -> None:
+def extract(
+    infilename: str, outfilename: str, start=0, end=-1, step=1, items=None
+) -> None:
     """Extract timesteps and/or items to a new dfs file
 
     Parameters
@@ -383,6 +440,8 @@ def extract(infilename: str, outfilename: str, start=0, end=-1, items=None) -> N
     end : int, float, str or datetime, optional
         end of extraction as either step, relative seconds
         or datetime/str, by default -1 (end of file)
+    step : int, optional
+        jump this many step, by default 1 (every step between start and end)
     items : int, list(int), str, list(str), optional
         items to be extracted to new file
 
@@ -391,24 +450,33 @@ def extract(infilename: str, outfilename: str, start=0, end=-1, items=None) -> N
     >>> extract('f_in.dfs0', 'f_out.dfs0', start='2018-1-1')
     >>> extract('f_in.dfs2', 'f_out.dfs2', end=-3)
     >>> extract('f_in.dfsu', 'f_out.dfsu', start=1800.0, end=3600.0)
+    >>> extract('f_hourly.dfsu', 'f_daily.dfsu', step=24)
     >>> extract('f_in.dfsu', 'f_out.dfsu', items=[2, 0])
     >>> extract('f_in.dfsu', 'f_out.dfsu', items="Salinity")
     >>> extract('f_in.dfsu', 'f_out.dfsu', end='2018-2-1 00:00', items="Salinity")
     """
-    dfs_i = DfsFileFactory.DfsGenericOpenEdit(infilename)
+    dfs_i = DfsFileFactory.DfsGenericOpenEdit(str(infilename))
 
-    is_dfsu_3d = dfs_i.ItemInfo[0].Name == "Z coordinate"
+    is_layered_dfsu = dfs_i.ItemInfo[0].Name == "Z coordinate"
 
     file_start_new, start_step, start_sec, end_step, end_sec = _parse_start_end(
         dfs_i, start, end
     )
-    item_numbers = _valid_item_numbers(dfs_i.ItemInfo, items)
+    timestep = _parse_step(dfs_i, step)
+    item_numbers = _valid_item_numbers(
+        dfs_i.ItemInfo, items, ignore_first=is_layered_dfsu
+    )
 
-    if is_dfsu_3d and 0 not in item_numbers:
+    if is_layered_dfsu:
+        item_numbers = [it + 1 for it in item_numbers]
         item_numbers.insert(0, 0)
 
     dfs_o = _clone(
-        infilename, outfilename, start_time=file_start_new, items=item_numbers
+        str(infilename),
+        str(outfilename),
+        start_time=file_start_new,
+        timestep=timestep,
+        items=item_numbers,
     )
 
     file_start_shift = 0
@@ -417,7 +485,7 @@ def extract(infilename: str, outfilename: str, start=0, end=-1, items=None) -> N
         file_start_shift = (file_start_new - file_start_orig).total_seconds()
 
     timestep_out = -1
-    for timestep in range(start_step, end_step):
+    for timestep in range(start_step, end_step, step):
         for item_out, item in enumerate(item_numbers):
             itemdata = dfs_i.ReadItemTimeStep((item + 1), timestep)
             time_sec = itemdata.Time
@@ -488,23 +556,23 @@ def _parse_start_end(dfs_i, start, end):
         end_sec = (end - file_start_datetime).total_seconds()
 
     if start_step < 0:
-        start_step = 0
-        warnings.warn("start cannot be before start of file")
+        raise ValueError(
+            f"start cannot be before start of file. start={start_step} is invalid"
+        )
 
     if start_sec < file_start_sec:
-        start_sec = file_start_sec
-        warnings.warn("start cannot be before start of file")
+        raise ValueError(
+            f"start cannot be before start of file start={start_step} is invalid"
+        )
 
     if (end_sec < start_sec) or (end_step < start_step):
         raise ValueError("end must be after start")
 
     if end_step > n_time_steps:
-        end_step = n_time_steps
-        warnings.warn("end cannot be after end of file")
+        raise ValueError(f"end cannot be after end of file. end={end_step} is invalid.")
 
     if end_sec > file_end_sec:
-        end_sec = file_end_sec
-        warnings.warn("end cannot be after end of file")
+        raise ValueError(f"end cannot be after end of file. end={end_sec} is invalid.")
 
     file_start_new = None
     if dfs_i.FileInfo.TimeAxis.TimeAxisType == 3:
@@ -520,6 +588,19 @@ def _parse_start_end(dfs_i, start, end):
     return file_start_new, start_step, start_sec, end_step, end_sec
 
 
+def _parse_step(dfs_i, step):
+    """Helper function for parsing step argument"""
+    if step == 1:
+        timestep = None
+    elif dfs_i.FileInfo.TimeAxis.TimeAxisType == 3:
+        timestep = dfs_i.FileInfo.TimeAxis.TimeStep * step
+    elif dfs_i.FileInfo.TimeAxis.TimeAxisType == 4:
+        timestep = None
+    else:
+        raise ValueError("TimeAxisType not supported")
+    return timestep
+
+
 def avg_time(infilename: str, outfilename: str, skipna=True):
     """Create a temporally averaged dfs file
 
@@ -552,7 +633,7 @@ def avg_time(infilename: str, outfilename: str, skipna=True):
         indata = indatatime.Data
         has_value = indata != deletevalue
         indata[~has_value] = np.nan
-        outdatalist.append(indata.astype(np.float64))
+        outdatalist.append(indata.astype(np.float32))
         step0 = np.zeros_like(indata, dtype=np.int32)
         step0[has_value] = 1
         steps_list.append(step0)
@@ -634,8 +715,8 @@ def quantile(
 
     qvec = [q] if np.isscalar(q) else q
     qtxt = [f"Quantile {q}" for q in qvec]
-    core_iteminfos = [dfs_i.ItemInfo[i] for i in item_numbers]
-    items = _get_repeated_items(core_iteminfos, prefixes=qtxt)
+    core_items = [dfs_i.ItemInfo[i] for i in item_numbers]
+    items = _get_repeated_items(core_items, prefixes=qtxt)
 
     if is_dfsu_3d:
         items.insert(0, dfs_i.ItemInfo[0])
