@@ -1,12 +1,10 @@
 import os
-from pathlib import Path
 import numpy as np
 import pandas as pd
 import datetime
 import mikeio
 from mikeio.dfs0 import Dfs0
-from mikeio.eum import TimeStepUnit, EUMType, EUMUnit, ItemInfo
-from datetime import timedelta
+from mikeio.eum import EUMType, EUMUnit, ItemInfo
 
 import pytest
 
@@ -31,111 +29,51 @@ def test_repr_equidistant():
     assert "NonEquidistant" not in text
 
 
-def test_simple_write(tmpdir):
-
-    filename = os.path.join(tmpdir.dirname, "simple.dfs0")
-
-    data = []
-
-    nt = 100
-    d = np.random.random([nt])
-    data.append(d)
-
-    dfs = Dfs0()
-
-    dfs.write(filename=filename, data=data)
-
-    assert os.path.exists(filename)
-
-    filepath = Path(filename)
-
-    filepath.unlink()  # Remove file
-
-    dfs.write(filepath, data=data)
-
-    assert filepath.exists
-
-
 def test_write_float(tmpdir):
 
     filename = os.path.join(tmpdir.dirname, "simple_float.dfs0")
 
-    data = []
-
     nt = 100
-    d = np.random.random([nt]).astype(np.float32)
-    data.append(d)
 
-    dfs = Dfs0()
+    da = mikeio.DataArray(
+        data=np.random.random([nt]).astype(np.float32),
+        time=pd.date_range("2000", periods=nt, freq="H"),
+    )
 
-    dfs.write(filename=filename, data=data)
+    da.to_dfs(filename)
 
     assert os.path.exists(filename)
 
 
 def test_write_double(tmpdir):
 
-    filename = os.path.join(tmpdir.dirname, "simple_double.dfs0")
-
-    data = []
+    filename = os.path.join(tmpdir.dirname, "simple_float.dfs0")
 
     nt = 100
-    d = np.random.random([nt])
-    data.append(d)
 
-    dfs = Dfs0()
+    da = mikeio.DataArray(
+        data=np.random.random([nt]).astype(np.float32),
+        time=pd.date_range("2000", periods=nt, freq="H"),
+    )
 
-    dfs.write(filename=filename, data=data, dtype=np.float64)
+    da.to_dfs(filename, dtype=np.float64)
 
     assert os.path.exists(filename)
 
 
 def test_write_int_not_possible(tmpdir):
 
-    filename = os.path.join(tmpdir.dirname, "simple_double.dfs0")
-
-    data = []
+    filename = os.path.join(tmpdir.dirname, "simple_float.dfs0")
 
     nt = 100
-    d = np.random.random([nt])
-    data.append(d)
 
-    dfs = Dfs0()
+    da = mikeio.DataArray(
+        data=np.random.random([nt]).astype(np.float32),
+        time=pd.date_range("2000", periods=nt, freq="H"),
+    )
 
     with pytest.raises(TypeError):
-        dfs.write(filename=filename, data=data, dtype=np.int32)
-
-
-def test_write_2darray(tmpdir):
-    """
-    Data should be ideally be supplied as a list of 1d timeseries.
-    But if I supply a 2d array (nitems, nt), that could work as well
-    """
-
-    filename = os.path.join(tmpdir.dirname, "from2darray.dfs0")
-
-    nt = 100
-    nitems = 2
-    data = np.random.random([nitems, nt])
-
-    dfs = Dfs0()
-
-    dfs.write(filename=filename, data=data)
-
-    assert os.path.exists(filename)
-
-    dfsnew = Dfs0(filename)
-    assert dfsnew.n_items == nitems
-
-    ds = dfsnew.read()
-    assert ds[0].shape == (nt,)
-
-    filepath = Path(filename)
-    dfsnew = Dfs0(filepath)
-    assert dfsnew.n_items == nitems
-
-    ds = dfsnew.read()
-    assert ds[0].shape == (nt,)
+        da.to_dfs(filename, dtype=np.int32)
 
 
 def test_read_units_write_new(tmpdir):
@@ -153,8 +91,7 @@ def test_read_units_write_new(tmpdir):
     newdfs = Dfs0(tmpfile)
     ds2 = newdfs.read()
 
-    assert ds2.items[0].type == ds.items[0].type
-    assert ds2.items[0].unit == ds.items[0].unit
+    assert ds2.items[0] == ds.items[0]
 
 
 def test_read_start_end_time():
@@ -168,25 +105,26 @@ def test_read_start_end_time():
     assert dfs.end_time == ds.end_time
 
 
-def test_multiple_write(tmpdir):
+def test_read_all_time_steps_without_reading_items():
 
-    filename = os.path.join(tmpdir.dirname, "random.dfs0")
+    dfs0file = r"tests/testdata/random.dfs0"
+    dfs = mikeio.open(dfs0file)
+    assert isinstance(dfs.time, pd.DatetimeIndex)
+    assert len(dfs.time) == 1000
 
-    data = []
+def test_items_dataframe():
+    dfs = mikeio.open("tests/testdata/random.dfs0")
+    df = dfs.items.to_dataframe()
+    assert "name" in df.columns
+    assert "type" in df.columns # or EUMType ?
+    assert df.type.iloc[1] == "Water_Level" # Is this the correct way to show it?
 
-    nt = 10
-    d1 = np.zeros(nt)
-    data.append(d1)
-    d2 = np.ones(nt)
-    data.append(d2)
 
-    items = [ItemInfo("Zeros"), ItemInfo("Ones")]
-
-    dfs = Dfs0()
-
-    dfs.write(filename=filename, data=data, items=items, title="Zeros and ones")
-
-    assert os.path.exists(filename)
+def test_read_all_time_steps_without_reading_items_neq():
+    dfs0file = r"tests/testdata/da_diagnostic.dfs0"
+    dfs = mikeio.open(dfs0file)
+    assert isinstance(dfs.time, pd.DatetimeIndex)
+    assert len(dfs.time) == 744
 
 
 def test_write_equidistant_calendar(tmpdir):
@@ -283,35 +221,32 @@ def test_read_dfs0_delete_value_conversion():
     dfs = Dfs0(dfs0file)
     ds = dfs.read()
 
-    assert np.isnan(ds.data[3][1])
+    assert np.isnan(ds[3].values[1])
 
     dfs0file = r"tests/testdata/random.dfs0"
 
     dfs = Dfs0(dfs0file)
     ds = dfs.read()
 
-    assert np.isnan(ds.data[0][2])
+    assert np.isnan(ds[0].values[2])
 
 
 def test_read_dfs0_small_value_not_delete_value(tmpdir):
 
     filename = os.path.join(tmpdir.dirname, "small.dfs0")
-
-    data = []
-
     d = np.array([0.0, 0.0000001, -0.0001])
-
     assert np.isclose(d, -1e-35, atol=1e-33).any()
-    data.append(d)
 
-    dfs = Dfs0()
+    da = mikeio.DataArray(
+        data=d,
+        time=pd.date_range("2000", periods=len(d), freq="H"),
+    )
 
-    dfs.write(filename=filename, data=data)
+    da.to_dfs(filename)
 
-    dfs = Dfs0(filename)
-    ds = dfs.read()
+    ds = mikeio.read(filename)
 
-    assert not np.isnan(ds.data[0]).any()
+    assert not np.isnan(ds[0].to_numpy()).any()
 
 
 def test_write_from_data_frame(tmpdir):
@@ -356,12 +291,42 @@ def test_write_from_data_frame_monkey_patched(tmpdir):
     ds = mikeio.read(filename)
 
     assert len(ds.items) == 5
-    assert ds.items[0].type == EUMType.Concentration
-    assert ds.items[0].unit == EUMUnit.gram_per_meter_pow_3
-    assert np.isnan(ds["Average"][3])
+    assert ds[0].type == EUMType.Concentration
+    assert ds[0].unit == EUMUnit.gram_per_meter_pow_3
+    assert np.isnan(ds["Average"].to_numpy()[3])
     assert ds.time[0].year == 1958
 
 
+def test_write_dataframe_different_eum_types_to_dfs0(tmpdir):
+
+    time = pd.DatetimeIndex(["2001-01-01", "2001-01-01 01:00", "2001-01-01 01:10"])
+
+    df = pd.DataFrame(
+        {"flow": np.array([1, np.nan, 2]), "level": np.array([2, 3.0, -1.3])}
+    )
+    df.index = time
+
+    dfr = df.resample("5min").mean().fillna(0.0)  # .interpolate()
+
+    filename = os.path.join(tmpdir.dirname, "dataframe.dfs0")
+
+    dfr.to_dfs0(
+        filename,
+        items=[
+            mikeio.ItemInfo("Flow", itemtype=mikeio.EUMType.Discharge),
+            mikeio.ItemInfo("Level", itemtype=mikeio.EUMType.Water_Level),
+        ],
+    )
+
+    ds = mikeio.read(filename)
+    assert ds.n_timesteps == 15
+    assert ds[0].type == mikeio.EUMType.Discharge
+    assert ds[1].type == mikeio.EUMType.Water_Level
+    assert len(ds) == 2
+    assert ds.end_time == dfr.index[-1]
+    assert ds.is_equidistant
+
+
 def test_write_from_pandas_series_monkey_patched(tmpdir):
 
     df = pd.read_csv(
@@ -382,9 +347,9 @@ def test_write_from_pandas_series_monkey_patched(tmpdir):
     ds = mikeio.read(filename)
 
     assert len(ds.items) == 1
-    assert ds.items[0].type == EUMType.Concentration
-    assert ds.items[0].unit == EUMUnit.gram_per_meter_pow_3
-    assert np.isnan(ds["Average"][3])
+    assert ds[0].type == EUMType.Concentration
+    assert ds[0].unit == EUMUnit.gram_per_meter_pow_3
+    assert np.isnan(ds["Average"].to_numpy()[3])
     assert ds.time[0].year == 1958
 
 
@@ -425,7 +390,7 @@ def test_read_dfs0_single_item():
     dfs = Dfs0(dfs0file)
     ds = dfs.read([1])
 
-    assert len(ds.data) == 1
+    assert len(ds.to_numpy()) == 1
 
 
 def test_read_dfs0_single_item_named_access():
@@ -434,9 +399,8 @@ def test_read_dfs0_single_item_named_access():
 
     dfs = Dfs0(dfs0file)
     res = dfs.read(items=[1])
-    data = res.data
 
-    assert len(data) == 1
+    assert len(res.to_numpy()) == 1
 
 
 def test_read_dfs0_temporal_subset():
@@ -444,7 +408,7 @@ def test_read_dfs0_temporal_subset():
     dfs0file = r"tests/testdata/random.dfs0"
 
     dfs = Dfs0(dfs0file)
-    ds = dfs.read(time_steps=[1, 2])
+    ds = dfs.read(time=[1, 2])
 
     assert len(ds.time) == 2
     assert ds.time[0].strftime("%H") == "05"
@@ -455,7 +419,7 @@ def test_read_non_eq_dfs0__temporal_subset():
     dfs0file = r"tests/testdata/da_diagnostic.dfs0"
 
     dfs = Dfs0(dfs0file)
-    ds = dfs.read(time_steps="2017-10-27 01:00,2017-10-27 02:00")
+    ds = dfs.read(time="2017-10-27 01:00,2017-10-27 02:00")
 
     assert len(ds.time) == 7
 
@@ -466,7 +430,7 @@ def test_read_dfs0_single_item_read_by_name():
 
     dfs = Dfs0(dfs0file)
     res = dfs.read(["NotFun", "VarFun01"])  # reversed order compare to original file
-    data = res.data
+    data = res.to_numpy()
 
     assert len(data) == 2
     assert res.items[0].name == "NotFun"
@@ -491,7 +455,7 @@ def test_read_dfs0_to_matrix():
     dfs = Dfs0(dfs0file)
     ds = dfs.read()
 
-    assert len(ds.data) == 2
+    assert len(ds.to_numpy()) == 2
 
 
 def test_write_data_with_missing_values(tmpdir):
@@ -502,21 +466,21 @@ def test_write_data_with_missing_values(tmpdir):
     ds = dfs.read()
 
     # Do something with the data
-    ds.data[0] = np.zeros_like(ds.data[0])
-    ds.data[1] = np.ones_like(ds.data[0])
+    ds[0].values = np.zeros_like(ds[0].values)
+    ds[1].values = np.ones_like(ds[0].values)
 
     # Add some NaNs
-    ds.data[1][0:10] = np.nan
+    ds[1].values[0:10] = np.nan
 
     # Overwrite the file
     dfs.write(tmpfile, ds)
 
     # Write operation does not modify the data
-    assert np.isnan(ds.data[1][1])
+    assert np.isnan(ds[1].values[1])
 
     moddfs = Dfs0(tmpfile)
     modified = moddfs.read()
-    assert np.isnan(modified.data[1][5])
+    assert np.isnan(modified[1].values[5])
 
 
 def test_read_relative_time_axis():
@@ -607,9 +571,9 @@ def test_write_from_pandas_series_monkey_patched_data_value_not_default(tmpdir):
     ds = mikeio.read(filename)
 
     assert len(ds.items) == 1
-    assert ds.items[0].type == EUMType.Concentration
-    assert ds.items[0].unit == EUMUnit.gram_per_meter_pow_3
-    assert np.isnan(ds["Average"][3])
+    assert ds[0].type == EUMType.Concentration
+    assert ds[0].unit == EUMUnit.gram_per_meter_pow_3
+    assert np.isnan(ds["Average"].to_numpy()[3])
     assert ds.time[0].year == 1958
     assert ds.items[0].data_value_type == 3
 
@@ -641,8 +605,49 @@ def test_write_from_data_frame_monkey_patched_data_value_not_default(tmpdir):
     ds = mikeio.read(filename)
 
     assert len(ds.items) == 5
-    assert ds.items[0].type == EUMType.Concentration
-    assert ds.items[0].unit == EUMUnit.gram_per_meter_pow_3
-    assert np.isnan(ds["Average"][3])
+    assert ds[0].type == EUMType.Concentration
+    assert ds[0].unit == EUMUnit.gram_per_meter_pow_3
+    assert np.isnan(ds["Average"].to_numpy()[3])
     assert ds.time[0].year == 1958
     assert ds.items[0].data_value_type == 3
+
+
+def test_read_write_eum(tmp_path):
+
+    ds = mikeio.read("tests/testdata/waterlevel_viken.dfs0")
+    assert ds["ST 2: WL (m)"].type == EUMType.Water_Level
+    assert ds["ST 2: WL (m)"].unit == EUMUnit.meter
+
+    outfilename = tmp_path / "same_same.dfs0"
+
+    ds.to_dfs(outfilename)
+
+    ds2 = mikeio.read(outfilename)
+    assert ds2["ST 2: WL (m)"].type == EUMType.Water_Level
+    assert ds2["ST 2: WL (m)"].unit == EUMUnit.meter
+
+
+def test_read_write_single_step(tmp_path):
+
+    ds = mikeio.read("tests/testdata/waterlevel_viken.dfs0", time=-1)
+    outfilename = tmp_path / "last_step.dfs0"
+    ds.to_dfs(outfilename)
+
+    dsnew = mikeio.read(outfilename)
+    assert dsnew.n_timesteps == 1
+    assert dsnew[0].to_numpy() == pytest.approx(-0.08139999955892563)
+
+
+def test_read_write_single_step_to_dataframe(tmp_path):
+    ds = mikeio.read("tests/testdata/da_diagnostic.dfs0", time=1)
+    df = ds.to_dataframe()
+    assert df.shape[0] == 1
+    assert df.iloc[0, 0] == pytest.approx(1.81134)
+    assert np.isnan(df.iloc[0, 3])
+
+
+def test_read_dfs0_with_many_items():
+
+    ds = mikeio.read("tests/testdata/many_items.dfs0")
+
+    assert ds.n_items == 800
