@@ -1,6 +1,7 @@
 import os
 from shutil import copyfile
 import numpy as np
+import pandas as pd
 import mikeio
 from mikeio import Dfsu
 from mikeio import generic
@@ -18,9 +19,9 @@ def test_add_constant(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org.data[0][0]
+    orgvalue = org[0].values[0]
     expected = orgvalue + 100.0
-    scaledvalue = scaled.data[0][0]
+    scaledvalue = scaled[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -34,9 +35,9 @@ def test_multiply_constant(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org.data[0][0]
+    orgvalue = org[0].values[0]
     expected = orgvalue * 1.5
-    scaledvalue = scaled.data[0][0]
+    scaledvalue = scaled[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -50,14 +51,14 @@ def test_multiply_constant_single_item_number(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue_speed = org.data[0][0][0]
+    orgvalue_speed = org[0].values[0][0]
     expected_speed = orgvalue_speed * 1.5
-    scaledvalue_speed = scaled.data[0][0][0]
+    scaledvalue_speed = scaled[0].values[0][0]
     assert scaledvalue_speed == pytest.approx(expected_speed)
 
-    orgvalue_dir = org.data[1][0][0]
+    orgvalue_dir = org[1].values[0, 0]
     expected_dir = orgvalue_dir
-    scaledvalue_dir = scaled.data[1][0][0]
+    scaledvalue_dir = scaled[1].values[0, 0]
     assert scaledvalue_dir == pytest.approx(expected_dir)
 
 
@@ -71,14 +72,14 @@ def test_multiply_constant_single_item_name(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue_speed = org["Wind speed"][0, 0]
+    orgvalue_speed = org["Wind speed"].to_numpy()[0, 0]
     expected_speed = orgvalue_speed * 1.5
-    scaledvalue_speed = scaled["Wind speed"][0, 0]
+    scaledvalue_speed = scaled["Wind speed"].to_numpy()[0, 0]
     assert scaledvalue_speed == pytest.approx(expected_speed)
 
-    orgvalue_dir = org["Wind direction"][0, 0]
+    orgvalue_dir = org["Wind direction"].to_numpy()[0, 0]
     expected_dir = orgvalue_dir
-    scaledvalue_dir = scaled["Wind direction"][0, 0]
+    scaledvalue_dir = scaled["Wind direction"].to_numpy()[0, 0]
     assert scaledvalue_dir == pytest.approx(expected_dir)
 
 
@@ -92,13 +93,13 @@ def test_diff_itself(tmpdir):
 
     org = mikeio.read(infilename_1)
 
-    assert np.isnan(org["Elevation"][0][0, -1])
+    assert np.isnan(org["Elevation"].to_numpy()[0, -1, -1])
 
     diffed = mikeio.read(outfilename)
 
-    diffedvalue = diffed["Elevation"][0, 0, 0]
+    diffedvalue = diffed["Elevation"].to_numpy()[0, 0, 0]
     assert diffedvalue == pytest.approx(0.0)
-    assert np.isnan(diffed["Elevation"][0][0, -1])
+    assert np.isnan(diffed["Elevation"].to_numpy()[0, -1, -1])
 
 
 def test_sum_itself(tmpdir):
@@ -111,11 +112,9 @@ def test_sum_itself(tmpdir):
 
     org = mikeio.read(infilename_1)
 
-    assert np.isnan(org["Elevation"][0][0, -1])
-
     summed = mikeio.read(outfilename)
 
-    assert np.isnan(summed["Elevation"][0][0, -1])
+    assert np.isnan(summed["Elevation"].to_numpy()[0][-1, -1])
 
 
 def test_add_constant_delete_values_unchanged(tmpdir):
@@ -128,14 +127,14 @@ def test_add_constant_delete_values_unchanged(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org["Elevation"][0, 0, 0]
-    scaledvalue = scaled["Elevation"][0, 0, 0]
+    orgvalue = org["Elevation"].to_numpy()[0, 0, 0]
+    scaledvalue = scaled["Elevation"].to_numpy()[0, 0, 0]
     assert scaledvalue == pytest.approx(orgvalue - 2.1)
 
-    orgvalue = org["Elevation"][0, 100, 0]
+    orgvalue = org["Elevation"].to_numpy()[0, 100, 0]
     assert np.isnan(orgvalue)
 
-    scaledvalue = scaled["Elevation"][0, 100, 0]
+    scaledvalue = scaled["Elevation"].to_numpy()[0, 100, 0]
     assert np.isnan(scaledvalue)
 
 
@@ -152,14 +151,14 @@ def test_multiply_constant_delete_values_unchanged_2(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org[item_name][0, 0, 0]
-    scaledvalue = scaled[item_name][0, 0, 0]
+    orgvalue = org[item_name].to_numpy()[0, 0, 0]
+    scaledvalue = scaled[item_name].to_numpy()[0, 0, 0]
     assert scaledvalue == pytest.approx(orgvalue * 1000.0)
 
-    orgvalue = org[item_name][0, 10, 0]
+    orgvalue = org[item_name].to_numpy()[0, -11, 0]
     assert np.isnan(orgvalue)
 
-    scaledvalue = scaled[item_name][0, 10, 0]
+    scaledvalue = scaled[item_name].to_numpy()[0, -11, 0]
     assert np.isnan(scaledvalue)
 
 
@@ -173,9 +172,9 @@ def test_linear_transform(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org.data[0][0]
+    orgvalue = org[0].values[0]
     expected = orgvalue * 1.5 - 20.0
-    scaledvalue = scaled.data[0][0]
+    scaledvalue = scaled[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -189,9 +188,9 @@ def test_linear_transform_dfsu(tmpdir):
 
     scaled = mikeio.read(outfilename)
 
-    orgvalue = org.data[0][0]
+    orgvalue = org[0].values[0]
     expected = orgvalue * 1.5 - 20.0
-    scaledvalue = scaled.data[0][0]
+    scaledvalue = scaled[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -206,9 +205,9 @@ def test_sum_dfsu(tmpdir):
 
     summed = mikeio.read(outfilename)
 
-    orgvalue = org.data[0][0]
+    orgvalue = org[0].values[0]
     expected = orgvalue * 2
-    scaledvalue = summed.data[0][0]
+    scaledvalue = summed[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -224,7 +223,7 @@ def test_diff_dfsu(tmpdir):
     diffed = mikeio.read(outfilename)
 
     expected = 0.0
-    scaledvalue = diffed.data[0][0]
+    scaledvalue = diffed[0].values[0]
     assert scaledvalue == pytest.approx(expected)
 
 
@@ -287,6 +286,69 @@ def test_concat_closes_files(tmpdir):
     # assert len(ds.time) == (5 * 48 + 1)
 
 
+def test_concat_keep(tmpdir):
+    """
+    test keep arguments of concatenation function
+    """
+    # added keep arguments to test
+    keep_args = ["first", "last"]
+
+    infiles = [
+        "tests/testdata/tide1.dfs1",
+        "tests/testdata/tide2_offset.dfs1",
+        "tests/testdata/tide4.dfs1",
+    ]
+    outfilename = os.path.join(tmpdir.dirname, "concat.dfs1")
+
+    # loop through keep args to test all
+    for keep_arg in keep_args:
+        mikeio.generic.concat(
+            infilenames=infiles, outfilename=outfilename, keep=keep_arg
+        )
+
+        # read outfile
+        dso = mikeio.read(outfilename)
+        df_o = pd.DataFrame(dso[0].to_numpy(), index=dso.time)
+
+        """ handle data (always keep two data in memory to compare in overlapping period) """
+        # !!! checks below solely consider one item !!
+        for i, infile in enumerate(infiles):
+            dsi = mikeio.read(infile)
+
+            # just store current (=last) data in first loop. no check required
+            if i == 0:
+                df_last = pd.DataFrame(dsi[0].to_numpy(), index=dsi.time)
+            else:
+                # move previously last to first
+                df_first = df_last.copy()
+                # overwrite last with new / current data
+                df_last = pd.DataFrame(dsi[0].to_numpy(), index=dsi.time)
+
+                # find overlapping timesteps via merge
+                overlap_idx = pd.merge(
+                    df_first, df_last, how="inner", left_index=True, right_index=True
+                ).index
+
+                # check if first and output are identical (results in dataframe with overlapping datetime index and bool values)
+                first_out = (
+                    (df_first.loc[overlap_idx] == df_o.loc[overlap_idx])
+                    .eq(True)
+                    .all()
+                    .all()
+                )
+                last_out = (
+                    (df_last.loc[overlap_idx] == df_o.loc[overlap_idx])
+                    .eq(True)
+                    .all()
+                    .all()
+                )
+
+                if keep_arg == "first":
+                    assert first_out, "overlap should be with first dataset"
+                elif keep_arg == "last":
+                    assert last_out, "overlap should be with last dataset"
+
+
 def test_extract_equidistant(tmpdir):
 
     infile = "tests/testdata/waves.dfs2"
@@ -298,17 +360,13 @@ def test_extract_equidistant(tmpdir):
     assert extracted.n_timesteps == orig.n_timesteps - 1
     assert extracted.time[0] == orig.time[1]
 
-    extract(infile, outfile, start=-23, end=1000)
-    orig = mikeio.read(infile)
-    extracted = mikeio.read(outfile)
-    assert extracted.n_timesteps == orig.n_timesteps
+    with pytest.raises(ValueError):
+        extract(infile, outfile, start=-23, end=1000)
 
-    extract(infile, outfile, start=-23.5)
-    orig = mikeio.read(infile)
-    extracted = mikeio.read(outfile)
-    assert extracted.n_timesteps == orig.n_timesteps
+    with pytest.raises(ValueError):
+        extract(infile, outfile, start=-23.5)
 
-    with pytest.raises(Exception):
+    with pytest.raises(ValueError):
         extract(infile, outfile, start=1000)
 
 
@@ -328,15 +386,14 @@ def test_extract_non_equidistant(tmpdir):
     assert extracted.time[0].hour == 0
     assert extracted.time[-1].minute == 0
 
-    extract(infile, outfile, start=1800.0, end=237981231.232)
-    extracted = mikeio.read(outfile)
-    assert extracted.n_timesteps > 0
+    with pytest.raises(ValueError):
+        extract(infile, outfile, start=1800.0, end=237981231.232)
 
     extract(infile, outfile, "2017-10-27,2017-10-28")
     extracted = mikeio.read(outfile)
     assert extracted.n_timesteps > 0
 
-    with pytest.raises(Exception):
+    with pytest.raises(ValueError):
         extract(infile, outfile, start=7200.0, end=1800.0)
 
 
@@ -349,6 +406,33 @@ def test_extract_relative_time_axis(tmpdir):
         extract(infile, outfile, start=0, end=4)
 
 
+def test_extract_step_equidistant(tmpdir):
+
+    infile = "tests/testdata/tide1.dfs1"  # 30min
+    outfile = os.path.join(tmpdir.dirname, "tide1_6hour.dfs1")
+
+    extract(infile, outfile, step=12)
+    orig = mikeio.read(infile)
+    extracted = mikeio.read(outfile)
+    assert extracted.n_timesteps == 9
+    assert extracted.timestep == 6 * 3600
+    assert extracted.time[0] == orig.time[0]
+    assert extracted.time[-1] == orig.time[-1]
+
+
+def test_extract_step_non_equidistant(tmpdir):
+
+    infile = "tests/testdata/da_diagnostic.dfs0"
+    outfile = os.path.join(tmpdir.dirname, "diagnostic_step3.dfs0")
+
+    extract(infile, outfile, step=3)
+    orig = mikeio.read(infile)
+    extracted = mikeio.read(outfile)
+    assert extracted.n_timesteps * 3 == orig.n_timesteps
+    assert extracted.time[0] == orig.time[0]
+    assert extracted.time[-1] == orig.time[-3]
+
+
 def test_extract_items(tmpdir):
 
     # This is a Dfsu 3d file (i.e. first item is Z coordinate)
@@ -357,26 +441,26 @@ def test_extract_items(tmpdir):
 
     extract(infile, outfile, items="Temperature")
     extracted = mikeio.read(outfile)
-    assert extracted.n_items == 2
-    assert extracted.items[0].name == "Z coordinate"
+    assert extracted.n_items == 1
+    assert extracted.items[0].name == "Temperature"
 
-    extract(infile, outfile, items=[0, 2])
+    extract(infile, outfile, items=[1])
     extracted = mikeio.read(outfile)
-    assert extracted.n_items == 2
-    assert extracted.items[1].name == "Salinity"
+    assert extracted.n_items == 1
+    assert extracted.items[0].name == "Salinity"
 
     extract(infile, outfile, items=range(0, 2))  # [0,1]
     extracted = mikeio.read(outfile)
     assert extracted.n_items == 2
-    assert extracted.items[1].name == "Temperature"
+    assert extracted.items[0].name == "Temperature"
 
-    extract(infile, outfile, items=["Salinity", 1])
+    extract(infile, outfile, items=["Salinity", 0])
     extracted = mikeio.read(outfile)
-    assert extracted.n_items == 3  # Z coordinate item is always included
+    assert extracted.n_items == 2
 
     with pytest.raises(Exception):
         # must be unique
-        extract(infile, outfile, items=["Salinity", 2])
+        extract(infile, outfile, items=["Salinity", 1])
 
     with pytest.raises(Exception):
         # no negative numbers
@@ -400,7 +484,7 @@ def test_time_average(tmpdir):
     assert org.time[0] == averaged.time[0]
     assert org.shape[1] == averaged.shape[1]
     assert averaged.shape[0] == 1
-    assert np.allclose(org.mean(axis=0)[0], averaged[0])
+    assert np.allclose(org.mean(axis=0)[0].to_numpy(), averaged[0].to_numpy())
 
 
 def test_time_average_dfsu_3d(tmpdir):
@@ -408,8 +492,8 @@ def test_time_average_dfsu_3d(tmpdir):
     outfilename = os.path.join(tmpdir, "oresund_sigma_z_avg.dfsu")
     avg_time(infilename, outfilename)
 
-    org = Dfsu(infilename)
-    averaged = Dfsu(outfilename)
+    org = mikeio.open(infilename)
+    averaged = mikeio.open(outfilename)
 
     assert averaged.n_timesteps == 1
     assert org.start_time == averaged.start_time
@@ -428,10 +512,10 @@ def test_time_average_deletevalues(tmpdir):
     assert all([a == b for a, b in zip(org.items, averaged.items)])
     assert org.time[0] == averaged.time[0]
     assert org.shape[1] == averaged.shape[1]
-    nan1 = np.isnan(org[0])
-    nan2 = np.isnan(averaged[0])
+    nan1 = np.isnan(org[0].to_numpy())
+    nan2 = np.isnan(averaged[0].to_numpy())
     assert np.all(nan1 == nan2)
-    assert np.allclose(org[0][~nan1], averaged[0][~nan2])
+    assert np.allclose(org[0].to_numpy()[~nan1], averaged[0].to_numpy()[~nan2])
 
 
 def test_quantile_dfsu(tmpdir):
@@ -443,7 +527,7 @@ def test_quantile_dfsu(tmpdir):
     org = mikeio.read(infilename).quantile(q=0.1, axis=0)
     q10 = mikeio.read(outfilename)
 
-    assert np.allclose(org[0], q10[0])
+    assert np.allclose(org[0].to_numpy(), q10[0].to_numpy())
 
 
 def test_quantile_dfsu_buffer_size(tmpdir):
@@ -455,7 +539,7 @@ def test_quantile_dfsu_buffer_size(tmpdir):
     org = mikeio.read(infilename).quantile(q=0.1, axis=0)
     q10 = mikeio.read(outfilename)
 
-    assert np.allclose(org[0], q10[0])
+    assert np.allclose(org[0].to_numpy(), q10[0].to_numpy())
 
 
 def test_quantile_dfs2(tmpdir):
@@ -467,7 +551,7 @@ def test_quantile_dfs2(tmpdir):
     org = mikeio.read(infilename).quantile(q=0.9, axis=0)
     q90 = mikeio.read(outfilename)
 
-    assert np.allclose(org[0], q90[0])
+    assert np.allclose(org[0].to_numpy(), q90[0].to_numpy())
 
 
 def test_quantile_dfs0(tmpdir):
@@ -479,7 +563,7 @@ def test_quantile_dfs0(tmpdir):
     org = mikeio.read(infilename).quantile(q=[0.01, 0.5], axis=0)
     qnt = mikeio.read(outfilename)
 
-    assert np.allclose(org[0], qnt[0])
+    assert np.allclose(org[0].to_numpy(), qnt[0].to_numpy())
     # assert np.allclose(org[5], qnt[5])
 
 
@@ -488,5 +572,17 @@ def test_quantile_dfsu_3d(tmpdir):
     outfilename = os.path.join(tmpdir, "oresund_sigma_z_avg.dfsu")
     generic.quantile(infilename, outfilename, q=[0.1, 0.9])
 
-    qd = Dfsu(outfilename)
+    qd = mikeio.open(outfilename)
     assert qd.n_timesteps == 1
+
+
+def test_dfs_ext_capitalisation(tmpdir):
+    filename = os.path.join("tests", "testdata", "waves2.DFS0")
+    ds = mikeio.open(filename)
+    ds = mikeio.read(filename)
+    ds.to_dfs(os.path.join(tmpdir, "void.DFS0"))
+    filename = os.path.join("tests", "testdata", "odense_rough2.MESH")
+    ds = mikeio.open(filename)
+    filename = os.path.join("tests", "testdata", "oresund_vertical_slice2.DFSU")
+    ds = mikeio.open(filename)
+    assert True
