@@ -67,7 +67,7 @@ def test_get_land_node_coordinates(tri_mesh):
 
     msh = tri_mesh
 
-    nc = msh.node_coordinates[msh.codes == 1]
+    nc = msh.node_coordinates[msh.geometry.codes == 1]
 
     assert nc.shape == (134, 3)
 
@@ -91,23 +91,22 @@ def test_set_z(tri_mesh):
 
 
 def test_set_codes(tri_mesh):
-    filename = os.path.join("tests", "testdata", "odense_rough.mesh")
     msh = tri_mesh
-    codes = msh.codes
-    assert msh.codes[2] == 2
+    codes = msh.geometry.codes
+    assert msh.geometry.codes[2] == 2
     codes[codes == 2] = 7  # work directly on reference
 
-    assert msh.codes[2] == 7
+    assert msh.geometry.codes[2] == 7
 
-    new_codes = msh.codes.copy()
+    new_codes = msh.geometry.codes.copy()
     new_codes[new_codes == 7] = 9
-    msh.codes = new_codes  # assign from copy
+    msh.geometry.codes = new_codes  # assign from copy
 
-    assert msh.codes[2] == 9
+    assert msh.geometry.codes[2] == 9
 
     with pytest.raises(ValueError):
         # not same length
-        msh.codes = codes[0:4]
+        msh.geometry.codes = codes[0:4]
 
 
 def test_write(tri_mesh, tmpdir):
